package HTML::StyleStream;
 
# $Id: StyleStream.pm,v 1.11 1998/11/25 17:43:37 emanuel Exp $
 
use strict;
use vars qw($VERSION @ISA $AUTOLOAD);
use Carp;
use HTML::Parser;
use HTML::Entities (); 

$VERSION = join '', q$Revision: 1.11 $ =~ /([\d\.]+)/;
@ISA = qw(HTML::Parser);

### Public Constructor ###
 
sub new {
  my($proto, %args) = @_;
  my $class = ref($proto) || $proto;
  my $self = $class->SUPER::new;
 
  my $style = exists $args{style} ? $args{style} 
                                  : "HTML::StyleStream::Plain";

  if (ref $style) {
    $self->{style} = $style;
  } else {
    my $require = "$style.pm";
    $require =~ s|::|/|g;
 
    $self->{style} = eval {
      require $require;
      new $style(%args);
    };
    if ($@) {
      warn $@ if $^W;
      return undef;
    }
    $self->{style} or return undef;
  }
  return $self;
}
 
### Public Methods ###
 
sub print {
  my $self = shift;
  $self->{style}->print(@_);
  return $self;
}
 
sub text {
  my $self = shift;
  if ((caller)[0] eq "HTML::Parser") { $self->{style}->print(@_) }
  else { $self->{style}->text(@_) }
  return $self;
}
 
sub entity {
  my $self = shift;
  $self->{style}->entity(@_);
  return $self;
}
 
sub nl {
  my $self = shift;
  $self->{style}->nl;
  return $self;
}
 
sub tag {
  my $self = shift;
 
  my $tag = uc shift;
  $tag =~ s|^/|_|;
  $tag =~ s|-|_|g;
 
  my @attr = ();
  while (@_) {
    push(@attr, uc shift);
    push(@attr, shift);
  }
 
  $self->{style}->$tag(@attr);
  return $self;
}

### Parser methods ###

sub declaration {
  my($self, $decl) = @_;
  $self->print("<!$decl>");
}

sub start {
  my($self, $tag, $attr, $attrseq, $origtext) = @_;
  $self->tag($tag, %$attr);
}

sub end {
  my($self, $tag) = @_;
  $self->tag("_$tag");
}

sub comment {
  my($self, $comment) = @_;
  $self->print("<!--$comment-->");
}

### Tie Methods ###

sub TIEHANDLE {
  my $class = shift;
  return $class->new(@_);
}

sub PRINT {
  my $self = shift;
  $self->parse(join("", @_));
}

sub PRINTF {
  my $self = shift;
  my $format = shift;
  $self->parse(sprintf($format, @_));
}

### Autoload HTML methods ###
 
sub AUTOLOAD {
  my $self = shift;
  my $name = $AUTOLOAD;
  $name =~ s/.*://;

  if ($name =~ /[a-z]/) {
    return $self->{style}->$name(@_);
  } else {
    my $text = ((@_ % 2 and shift) or undef);
    $self->{style}->$name(@_);
    if (defined $text) {
      $name = "_$name";
      $self->text($text)->$name();
    }
    return $self;
  }
}

### Destructor ###

sub DESTROY {}
 
1;
__END__

=head1 NAME

HTML::StyleStream -- Apply styles to HTML using re-write rules, and easily
output HTML from scripts.

=head1 SYNOPSIS

  # Constructor
  $sty = new HTML::StyleStream([style => 'style name' | $style_class] [,] [output => $handle] [...]);

  # HTML Output
  $sty->TAG([ATTR1=>'value1', ATTR2=>'value2', ...])
      ->text('Text to be escaped')
      ->_TAG;
      ->print('Verbatim text')
      ->nl;

  # Shortcut puts "text" between <TAG> ... </TAG> pair
  $sty->TAG("text", ATTR1=>'value1', ATTR2=>'value2', ...);

  # Alternate tag output
  $sty->tag('TAG', ATTR1=>'value1', ATTR2=>'value2', ...)
      ->tag('/TAG');

  # Parse HTML markup and apply style to it
  $sty->parse("... <TAG ATTR=value> HTML Markup </TAG> ...");

  # Tie a style to file handle and parse all HTML printed to it
  $sty = tie(*HANDLE, 'HTML::StyleStream' [, style => 'style name'] [, output => $fh]);
  print HANDLE "... <TAG ATTR=value> HTML Markup </TAG> ...";

=head1 DESCRIPTION

The use for this class is two-fold.  The first use is to easily output HTML
from perl programs.  The second and complementary use is to apply a "style" to
HTML.  A style is a series of rewrite rules.  For example, a rewrite rule
could be I<replace E<lt>H1E<gt> tags with E<lt>PE<gt>E<lt>FONT SIZE=+1 FACE=Gothic COLOR=redE<gt>>.
These styles can be applied to the output of your CGI programs, or it can
parse HTML and apply the style.  This makes it easy to keep more of your
layout information in one place and to change the appearance of your whole web
site without having to edit all your HTML files and CGI scripts.
Styles are defined using Perl's inheritance system.

=head1 CONSTRUCTOR

To get a new style stream object, use the constructor B<new>.
Arguments to the constructor are passed using "named arguments"
(i.e. a hash list).  I<style> specifies the document style class,
and defaults to "HTML::StyleStream::Plain".  I<style>
may also be a style object.  I<output> specifies
a reference to a file handle where the StyleStream will send its output
(in fact, I<output> may be any object with a B<print> method).  If not 
specified, output is sent to STDOUT.  All other arguments
are sent to the style class' constructor.  B<new> returns undef if the
style could not be loaded.
For example,

  $sty = new HTML::StyleStream( style => 'MyStyle', output => \*HANDLE ) or die "Cannot get MyStyle";

=head1 OUTPUT

=over 4

=item *

To output an HTML tag, invoke the method with the same name as the
tag on the StyleStream object.  Pass any attributes for the tag
as a hash list.  Prepend an underscore ('_') to the method name to
indicate an end tag.  For example,

  $sty->H1( ALIGN => 'center' );
  $sty->_H1;

outputs 

  <H1 ALIGN="center"></H1>

Any underscores, except the first, in the method name will be converted
to dashes on output.  There is currently no way to output a tag
containing a period ('.') using this technique.  Any letters in the
HTML output method names are capitalized.  All letters in the names
of the attributes must also be captilized.

=item *

Output text using the B<text> method.  This will escape any special
characters in the string(s).  Use B<print> to print text exactly
as-is.  For example,

  $sty->text("$a > 4 && $a < 8\n");
  $sty->print("Copyright &copy; 1998 Emanuel Borsboom");

outputs

  $a &gt; 4 &amp;&amp; $a &lt; 8
  Copyright &copy; 1998 Emanuel Borsboom

=item *

Output an HTML entity using B<entity>.  For example,

  $sty->entity('copy');

outputs "&copy;".  

=item *

Use B<nl> to output a newline character.

=item *

All of these methods return the object that they were invoked with,
which allows chaining.  For example,

  $sty->H1( ALIGN => 'center' )
      ->text("Hello, world")
      ->_H1;

=item *

As a shortcut, the above can also be written as

  $sty->H1("Hello, world", ALIGN => 'center')

If the first argument to a start tag is some text rather than an attribute,
the tag it will be expanded to surround the text with the start tag and the
matching end tag.

=item *

An alternate way to output a tag is using the B<tag> method.
The first argument is the tag, followed by a hash list with the
attributes.  The case of the tag and attribute names in
not significant.  For example,

  $sty->tag("h1", ALigN => 'center');

=back

=head1 STYLES

Styles are used to redefine existing HTML tags or to define your own.
This makes it easy set a consistent appearance among your stripts.

To define a document style, create a subclass of 
HTML::StyleStream::Plain.  To create or redefine a tag, write
a method with the same name as the tag which takes a hash list
containing the attributes as an argument.  I<All letters in
your tag and attribute names must be upper case!>  Your method
may invoke other tag methods and any of B<print>, B<text>, 
B<entity>, and B<nl>.  Indicate an end tag be prepending
an underscore ('_') to the method name.  This underscore will
be converted to a slash ('/') on output.  Any other underscores
will be converted to dashes ('-') on output.

Here is an example which defines a new document style called 
MyStyle which redefines H1 to be centered with a red font, 
and defines a new ASIDE tag:

  package MyStyle;

  @ISA = qw(HTML::StyleStream::Plain);

  sub H1 {
    my $self = shift;
    $self->SUPER::H1( ALIGN => 'center', @_ )
         ->FONT( COLOR => 'red' );
  }

  sub _H1 {
    shift->_FONT
         ->SUPER::_H1;
  }

  sub ASIDE {
    my($self, %attr) = @_;
    my $size = ($attr{SIZE} || '-1');
    $self->FONT( SIZE => $size )
         ->I;
  }

  sub _ASIDE {
    shift->_I
         ->_FONT;
  }

A couple things to notice:

=over 4

=item *

The redefined H1 invokes the superclass's (in this pase, plain) H1.
Be careful, because if you had written

  $self->H1( ALIGN => 'center', @_ )

instead, it would have caused infinite recursion.

=item *

The redefined H1 passes any attributes on the the superclass's.
Notice the order.  The way it is written, a subclass's style
or the user can override the new alignment by specifying a different
ALIGN attribute
because the user's attribute will come after the style's in the
hash list.  If it was written as

  $self->SUPER::H1( @_, ALIGN => 'center' )

subclasses and the user would not be able to override the style's
ALIGN attribute.

=back

Using I<MyStyle> defined above,

  $sty->H1->text("Hello, world")->_H1->nl
      ->ASIDE->text("Look behind you")->_ASIDE

outputs

  <H1 ALIGN=center><FONT COLOR="red">Hello, world</FONT></H1>
  <FONT SIZE="-1"><I>Look behind you</I></FONT>

=head1 PARSING

StyleStream can parse HTML markup and apply a style to it.
To do so, use the B<parse> method.  For example,

  $sty->parse("<H1>Hello, world</H1>\n<ASIDE>Look behind you</ASIDE>")

will output exactly the same as above.  In addition to B<parse>, 
StyleStream inherits B<eof> and B<parse_file> from HTML::Parser.
See HTML::Parser(3) for details.

Parsing may be safely interspersed with other output techniques,
though it would be wise to invoke B<eof> between B<parse> and any
other output methods.

=head1 CAPTURING OUTPUT

HTML::StyleStream can be tied to a file handle.  Anything that is 
printed to the file handle will be parsed as HTML as though B<parse>
were invoked.  This allows you to capture output and apply
the style to any HTML written.  Here's an example:

  $sty = tie *H, 'HTML::StyleStream', style => 'MyStyle';
  select H;
  print <<"--END--";
  <H1>Hello, world</H1>
  <ASIDE>Look behind you</ASIDE>
  --END--

When used with a package such as ePerl, this is an elegant
way of applying a style to the output of your script.

If you really want to capture STDOUT (rather than use the select hack),
be careful.  Writing this:

  $sty = tie *STDOUT, 'HTML::StyleStream', style => 'MyStyle';

will cause any output to loop back to itself (as soon as you print
to STDOUT, the style will be applied and output to STDOUT, which will apply
the style again, etc.).  Instead, use a class such as IO::Scalar (which
can be found on CPAN) to save the output, as follows:

  $output = '';
  $SH = new IO::Scalar \$output;
  $sty = tie *STDOUT, 'HTML::StyleStream', style => 'MyStyle', output => $SH;

  print <<"--END--";
  <H1>Hello, world</H1>
  <ASIDE>Look behind you</ASIDE>
  --END--

  undef $sty;
  untie *STDOUT;
  print $output;

You can still use the object returned by tie as a usual HTML::StyleStream
object.

=head1 ACKNOWLEDGEMENTS

Thanks to Eryq E<lt>eryq@zeegee.comE<gt> for writing HTML::Stream,
which was the inspiration for this class.

=head1 VERSION

$Id: StyleStream.pm,v 1.11 1998/11/25 17:43:37 emanuel Exp $

=head1 AUTHOR

Emanuel Borsboom E<lt>emanuel@zerius.comE<gt> July 1998

=head1 SEE ALSO

HTML::Parser, HTML::Stream, IO::Scalar, 
HTML::ePerl, HTML::Entities, http://zerius.com/emanuel/stylestream/

=cut
